# Copyright 2015 Rackspace Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import eventlet
import fixtures
import functools
import logging as pylogging
import platform
import sys
import time
from unittest import mock

from oslo_log import formatters
from oslo_log import log as logging
from oslotest import base
import testtools

from oslo_privsep import capabilities
from oslo_privsep import comm
from oslo_privsep import daemon
from oslo_privsep.tests import testctx


LOG = logging.getLogger(__name__)


def undecorated():
    pass


class TestException(Exception):
    pass


def get_fake_context(conf_attrs=None, **context_attrs):
    conf_attrs = conf_attrs or {}
    context = mock.NonCallableMock()
    context.conf.user = 42
    context.conf.group = 84
    context.conf.thread_pool_size = 10
    context.conf.capabilities = [
        capabilities.CAP_SYS_ADMIN,
        capabilities.CAP_NET_ADMIN,
    ]
    context.conf.logger_name = 'oslo_privsep.daemon'
    vars(context).update(context_attrs)  # type: ignore[attr-defined]
    vars(context.conf).update(conf_attrs)
    return context


@testctx.context.entrypoint
def logme(level, msg, exc_info=False):
    # We want to make sure we log everything from the priv side for
    # the purposes of this test, so force loglevel.
    LOG.logger.setLevel(logging.DEBUG)
    if exc_info:
        try:
            raise TestException('with arg')
        except TestException:
            LOG.log(level, msg, exc_info=True)
    else:
        LOG.log(level, msg)


@testctx.context.entrypoint
def raise_runtimeerror():
    raise RuntimeError()


class LogRecorder(pylogging.Formatter):
    def __init__(self, logs, *args, **kwargs):
        kwargs['validate'] = False
        super().__init__(*args, **kwargs)
        self.logs = logs

    def format(self, record):
        self.logs.append(copy.deepcopy(record))
        return super().format(record)


@testtools.skipIf(
    platform.system() != 'Linux', 'works only on Linux platform.'
)
class LogTest(testctx.TestContextTestCase):
    def test_priv_loglevel(self):
        logger = self.useFixture(fixtures.FakeLogger(level=logging.INFO))

        # These write to the log on the priv side
        logme(logging.DEBUG, 'test@DEBUG')
        logme(logging.WARNING, 'test@WARN')

        time.sleep(0.1)  # Hack to give logging thread a chance to run

        # logger.output is the resulting log on the unpriv side.
        # This should have been filtered based on (unpriv) loglevel.
        self.assertNotIn('test@DEBUG', logger.output)
        self.assertIn('test@WARN', logger.output)

    def test_record_data(self):
        logs: list[pylogging.LogRecord] = []
        # fixtures.FakeLogger accepts only a formatter class/function, not an
        # instance :(
        fmt = functools.partial(LogRecorder, logs)

        self.useFixture(
            fixtures.FakeLogger(
                level=logging.INFO,
                format='dummy',
                formatter=fmt,  # type: ignore[arg-type]
            )
        )

        try:
            logme(logging.WARNING, 'test with exc', exc_info=True)
        except Exception:  # noqa: S110
            pass

        time.sleep(0.1)  # Hack to give logging thread a chance to run

        self.assertEqual(1, len(logs))

        record = logs[0]
        self.assertIn('test with exc', record.getMessage())
        self.assertIsNone(record.exc_info)
        self.assertIn('TestException: with arg', record.exc_text)
        self.assertEqual(
            'PrivContext(cfg_section=privsep)', record.processName
        )
        self.assertIn('test_daemon.py', record.exc_text)
        self.assertEqual(logging.WARNING, record.levelno)
        self.assertEqual('logme', record.funcName)

    def test_format_record(self):
        logs: list[pylogging.LogRecord] = []
        # fixtures.FakeLogger accepts only a formatter class/function, not an
        # instance :(
        fmt = functools.partial(LogRecorder, logs)

        self.useFixture(
            fixtures.FakeLogger(
                level=logging.INFO,
                format='dummy',
                formatter=fmt,  # type: ignore[arg-type]
            )
        )

        logme(logging.WARNING, 'test with exc', exc_info=True)

        time.sleep(0.1)  # Hack to give logging thread a chance to run

        self.assertEqual(1, len(logs))

        record = logs[0]
        # Verify the log record can be formatted by ContextFormatter
        fake_config = mock.Mock(
            logging_default_format_string="NOCTXT: %(message)s"
        )
        formatter = formatters.ContextFormatter(config=fake_config)
        formatter.format(record)


@testtools.skipIf(
    platform.system() != 'Linux', 'works only on Linux platform.'
)
class LogTestDaemonTraceback(testctx.TestContextTestCase):
    def setUp(self):
        self.config_override = {'log_daemon_traceback': True}
        super().setUp()

    def test_record_daemon_traceback(self):
        self.privsep_conf.set_override(
            'log_daemon_traceback', True, group='privsep'
        )
        logs: list[pylogging.LogRecord] = []
        # fixtures.FakeLogger accepts only a formatter class/function, not an
        # instance :(
        fmt = functools.partial(LogRecorder, logs)

        self.useFixture(
            fixtures.FakeLogger(
                level=logging.INFO,
                format='dummy',
                formatter=fmt,  # type: ignore[arg-type]
            )
        )

        self.assertRaises(RuntimeError, raise_runtimeerror)
        time.sleep(0.1)  # Hack to give logging thread a chance to run

        self.assertEqual(1, len(logs))
        record = logs[0]
        self.assertIn('Privsep daemon traceback: ', record.getMessage())
        self.assertIsNone(record.exc_info)
        self.assertEqual('MainProcess', record.processName)
        self.assertEqual(logging.WARNING, record.levelno)


@testtools.skipIf(
    platform.system() != 'Linux', 'works only on Linux platform.'
)
class DaemonTest(base.BaseTestCase):
    @mock.patch('os.setuid')
    @mock.patch('os.setgid')
    @mock.patch('os.setgroups')
    @mock.patch('oslo_privsep.capabilities.set_keepcaps')
    @mock.patch('oslo_privsep.capabilities.drop_all_caps_except')
    def test_drop_privs(
        self,
        mock_dropcaps,
        mock_keepcaps,
        mock_setgroups,
        mock_setgid,
        mock_setuid,
    ):
        channel = mock.NonCallableMock()
        context = get_fake_context()

        manager = mock.Mock()
        manager.attach_mock(mock_setuid, "setuid")
        manager.attach_mock(mock_setgid, "setgid")
        expected_calls = [mock.call.setgid(84), mock.call.setuid(42)]

        d = daemon.Daemon(channel, context)
        d._drop_privs()

        mock_setuid.assert_called_once_with(42)
        mock_setgid.assert_called_once_with(84)
        mock_setgroups.assert_called_once_with([])

        self.assertEqual(expected_calls, manager.mock_calls)

        self.assertCountEqual(
            [mock.call(True), mock.call(False)], mock_keepcaps.mock_calls
        )

        mock_dropcaps.assert_called_once_with(
            {capabilities.CAP_SYS_ADMIN, capabilities.CAP_NET_ADMIN},
            {capabilities.CAP_SYS_ADMIN, capabilities.CAP_NET_ADMIN},
            [],
        )


@testtools.skipIf(
    platform.system() != 'Linux', 'works only on Linux platform.'
)
class WithContextTest(testctx.TestContextTestCase):
    def test_unexported(self):
        self.assertRaisesRegex(
            NameError,
            'undecorated not exported',
            testctx.context._wrap,
            undecorated,
        )


class ClientChannelTestCase(base.BaseTestCase):
    DICT = {
        'string_1': ('tuple_1', b'tuple_2'),
        b'byte_1': ['list_1', 'list_2'],
    }

    EXPECTED = {
        'string_1': ('tuple_1', b'tuple_2'),
        'byte_1': ['list_1', 'list_2'],
    }

    def setUp(self):
        super().setUp()
        context = get_fake_context()
        with (
            mock.patch.object(comm.ClientChannel, '__init__'),
            mock.patch.object(daemon._ClientChannel, 'exchange_ping'),
        ):
            self.client_channel = daemon._ClientChannel(mock.ANY, context)

    @mock.patch.object(daemon.LOG.logger, 'handle')
    def test_out_of_band_log_message(self, handle_mock):
        message = [comm.Message.LOG, self.DICT]
        self.assertEqual(self.client_channel.log, daemon.LOG)
        with (
            mock.patch.object(pylogging, 'makeLogRecord') as mock_make_log,
            mock.patch.object(
                daemon.LOG, 'isEnabledFor', return_value=True
            ) as mock_enabled,
        ):
            self.client_channel.out_of_band(message)
            mock_make_log.assert_called_once_with(self.EXPECTED)
            handle_mock.assert_called_once_with(mock_make_log.return_value)
            mock_enabled.assert_called_once_with(
                mock_make_log.return_value.levelno
            )

    def test_out_of_band_not_log_message(self):
        with mock.patch.object(daemon.LOG, 'warning') as mock_warning:
            self.client_channel.out_of_band([comm.Message.PING])
            mock_warning.assert_called_once()

    @mock.patch.object(daemon.logging, 'getLogger')  # type: ignore[attr-defined]
    @mock.patch.object(pylogging, 'makeLogRecord')
    def test_out_of_band_log_message_context_logger(
        self, make_log_mock, get_logger_mock
    ):
        logger_name = 'os_brick.privileged'
        context = get_fake_context(conf_attrs={'logger_name': logger_name})
        with (
            mock.patch.object(comm.ClientChannel, '__init__'),
            mock.patch.object(daemon._ClientChannel, 'exchange_ping'),
        ):
            channel = daemon._ClientChannel(mock.ANY, context)

        get_logger_mock.assert_called_once_with(logger_name)
        self.assertEqual(get_logger_mock.return_value, channel.log)

        message = [comm.Message.LOG, self.DICT]
        channel.out_of_band(message)

        make_log_mock.assert_called_once_with(self.EXPECTED)
        channel.log.isEnabledFor.assert_called_once_with(  # type: ignore[attr-defined]
            make_log_mock.return_value.levelno
        )
        channel.log.logger.handle.assert_called_once_with(  # type: ignore[attr-defined]
            make_log_mock.return_value
        )


class UnMonkeyPatch(base.BaseTestCase):
    def test_un_monkey_patch(self):
        self.assertFalse(
            any(
                eventlet.patcher.is_monkey_patched(eventlet_mod_name)
                for eventlet_mod_name in daemon.EVENTLET_MODULES
            )
        )

        eventlet.monkey_patch()
        self.assertTrue(
            any(
                eventlet.patcher.is_monkey_patched(eventlet_mod_name)
                for eventlet_mod_name in daemon.EVENTLET_MODULES
            )
        )

        daemon.un_monkey_patch()
        for eventlet_mod_name, func_modules in daemon.EVENTLET_LIBRARIES:
            if not eventlet.patcher.is_monkey_patched(eventlet_mod_name):
                continue

            for name, green_mod in func_modules():
                orig_mod = eventlet.patcher.original(name)
                patched_mod = sys.modules.get(name)
                for attr_name in green_mod.__patched__:
                    un_monkey_patched_attr = getattr(
                        patched_mod, attr_name, None
                    )
                    original_attr = getattr(orig_mod, attr_name, None)
                    self.assertEqual(un_monkey_patched_attr, original_attr)



import os.path


def load_tests(loader, tests, pattern):
    this_dir = os.path.dirname(__file__)
    new_tests = loader.discover(start_dir=this_dir, pattern=pattern)
    tests.addTests(new_tests)
    return tests



# Copyright 2019 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os
import time
import unittest

from oslo_config import fixture as config_fixture
from oslotest import base

from oslo_privsep import comm
from oslo_privsep import priv_context


test_context = priv_context.PrivContext(
    __name__,
    cfg_section='privsep',
    pypath=__name__ + '.test_context',
    capabilities=[],
)

test_context_with_timeout = priv_context.PrivContext(
    __name__,
    cfg_section='privsep',
    pypath=__name__ + '.test_context_with_timeout',
    capabilities=[],
    timeout=0.03,
)


@test_context.entrypoint
def sleep():
    # We don't want the daemon to be able to handle these calls too fast.
    time.sleep(0.001)


@test_context.entrypoint_with_timeout(0.03)
def sleep_with_timeout(long_timeout=0.04):
    time.sleep(long_timeout)
    return 42


@test_context_with_timeout.entrypoint
def sleep_with_t_context(long_timeout=0.04):
    time.sleep(long_timeout)
    return 42


@test_context.entrypoint
def one():
    return 1


@test_context.entrypoint
def logs():
    logging.warning('foo')  # noqa: LOG015


class TestDaemon(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        venv_path = os.environ['VIRTUAL_ENV']
        self.cfg_fixture = self.useFixture(config_fixture.Config())
        self.cfg_fixture.config(
            group='privsep',
            helper_command=f'sudo -E {venv_path}/bin/privsep-helper',
        )
        priv_context.init()

    def test_concurrency(self):
        # Throw a large number of simultaneous requests at the daemon to make
        # sure it can can handle them.
        for i in range(1000):
            sleep()
        # Make sure the daemon is still working
        self.assertEqual(1, one())

    def test_entrypoint_with_timeout(self):
        thread_pool_size = self.cfg_fixture.conf.privsep.thread_pool_size
        for _ in range(thread_pool_size + 1):
            self.assertRaises(comm.PrivsepTimeout, sleep_with_timeout)

    def test_entrypoint_with_timeout_pass(self):
        thread_pool_size = self.cfg_fixture.conf.privsep.thread_pool_size
        for _ in range(thread_pool_size + 1):
            res = sleep_with_timeout(0.01)
            self.assertEqual(42, res)

    def test_context_with_timeout(self):
        thread_pool_size = self.cfg_fixture.conf.privsep.thread_pool_size
        for _ in range(thread_pool_size + 1):
            self.assertRaises(comm.PrivsepTimeout, sleep_with_t_context)

    def test_context_with_timeout_pass(self):
        thread_pool_size = self.cfg_fixture.conf.privsep.thread_pool_size
        for _ in range(thread_pool_size + 1):
            res = sleep_with_t_context(0.01)
            self.assertEqual(42, res)

    def test_logging(self):
        logs()
        self.assertIn('foo', self.log_fixture.logger.output)


if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Rackspace Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Serialization/Deserialization for privsep.

The wire format is a stream of msgpack objects encoding primitive
python datatypes.  Msgpack 'raw' is assumed to be a valid utf8 string
(msgpack 2.0 'bin' type is used for bytes).  Python lists are
converted to tuples during serialization/deserialization.
"""

from __future__ import annotations

from collections.abc import Iterator
import datetime
import enum
import logging
import socket
import sys
import threading
from typing import Any

from typing_extensions import Self

import msgpack

from oslo_privsep._i18n import _
from oslo_utils import uuidutils

LOG = logging.getLogger(__name__)


@enum.unique
class Message(enum.IntEnum):
    """Types of messages sent across the communication channel"""

    PING = 1
    PONG = 2
    CALL = 3
    RET = 4
    ERR = 5
    LOG = 6


class PrivsepTimeout(Exception):
    pass


class Serializer:
    def __init__(self, writesock: socket.socket) -> None:
        self.writesock = writesock

    def send(self, msg: Any) -> None:
        buf = msgpack.packb(
            msg, use_bin_type=True, unicode_errors='surrogateescape'
        )
        self.writesock.sendall(buf)

    def close(self) -> None:
        # Hilarious. `socket._socketobject.close()` doesn't actually
        # call `self._sock.close()`.  Oh well, we really wanted a half
        # close anyway.
        self.writesock.shutdown(socket.SHUT_WR)


class Deserializer:
    def __init__(self, readsock: socket.socket) -> None:
        self.readsock = readsock
        self.unpacker = msgpack.Unpacker(
            use_list=False,
            raw=False,
            strict_map_key=False,
            unicode_errors='surrogateescape',
            # msgpack 1.0.0 set this value to 100MiB; raise it here for msgpack
            # < 1.0.0
            max_buffer_size=100 * 1024 * 1024,
        )

    def __iter__(self) -> Self:
        return self

    def __next__(self) -> Any:
        while True:
            try:
                return next(self.unpacker)
            except StopIteration:
                try:
                    buf = self.readsock.recv(4096)
                    if not buf:
                        raise
                    self.unpacker.feed(buf)
                except TimeoutError:
                    pass


class Future:
    """A very simple object to track the return of a function call"""

    def __init__(
        self, lock: threading.Lock, timeout: float | None = None
    ) -> None:
        self.condvar = threading.Condition(lock)
        self.error: BaseException | None = None
        self.data: Any = None
        self.timeout = timeout

    def set_result(self, data: Any) -> None:
        """Must already be holding lock used in constructor"""
        self.data = data
        self.condvar.notify()

    def set_exception(self, exc: BaseException) -> None:
        """Must already be holding lock used in constructor"""
        self.error = exc
        self.condvar.notify()

    def result(self) -> Any:
        """Must already be holding lock used in constructor"""
        before = datetime.datetime.now()
        if not self.condvar.wait(timeout=self.timeout):
            now = datetime.datetime.now()
            LOG.warning(
                'Timeout while executing a command, timeout: %s, '
                'time elapsed: %s',
                self.timeout,
                (now - before).total_seconds(),
            )
            return (
                Message.ERR.value,
                f'{PrivsepTimeout.__module__}.{PrivsepTimeout.__name__}',
                '',
            )
        if self.error is not None:
            raise self.error
        return self.data


class ClientChannel:
    def __init__(self, sock: socket.socket) -> None:
        self.running = False
        self.writer = Serializer(sock)
        self.lock = threading.Lock()
        self.reader_thread = threading.Thread(
            name='privsep_reader',
            target=self._reader_main,
            args=(Deserializer(sock),),
        )
        self.reader_thread.daemon = True
        self.outstanding_msgs: dict[str, Future] = {}

        self.reader_thread.start()

    def _reader_main(self, reader: Deserializer) -> None:
        """This thread owns and demuxes the read channel"""
        with self.lock:
            self.running = True
        for msg in reader:
            msgid, data = msg
            if msgid is None:
                self.out_of_band(data)
            else:
                with self.lock:
                    if msgid not in self.outstanding_msgs:
                        LOG.warning(
                            "msgid should be in oustanding_msgs, it is"
                            "possible that timeout is reached!"
                        )
                        continue
                    self.outstanding_msgs[msgid].set_result(data)

        # EOF.  Perhaps the privileged process exited?
        # Send an IOError to any oustanding waiting readers.  Assuming
        # the write direction is also closed, any new writes should
        # get an immediate similar error.
        LOG.debug('EOF on privsep read channel')

        exc = OSError(_('Premature eof waiting for privileged process'))
        with self.lock:
            for mbox in self.outstanding_msgs.values():
                mbox.set_exception(exc)
            self.running = False

    def out_of_band(self, msg: Any) -> None:
        """Received OOB message. Subclasses might want to override this."""
        pass

    def send_recv(self, msg: Any, timeout: float | None = None) -> Any:
        myid = uuidutils.generate_uuid()
        while myid in self.outstanding_msgs:
            LOG.warning("myid shoudn't be in outstanding_msgs.")
            myid = uuidutils.generate_uuid()
        future = Future(self.lock, timeout)

        with self.lock:
            self.outstanding_msgs[myid] = future
            try:
                self.writer.send((myid, msg))

                reply = future.result()
            except Exception:
                LOG.warning("Unexpected error: %s", sys.exc_info()[0])
                raise
            finally:
                del self.outstanding_msgs[myid]

        return reply

    def close(self) -> None:
        with self.lock:
            self.writer.close()

        self.reader_thread.join()


class ServerChannel:
    """Server-side twin to ClientChannel"""

    def __init__(self, sock: socket.socket) -> None:
        self.rlock = threading.Lock()
        self.reader_iter: Iterator[Any] = iter(Deserializer(sock))
        self.wlock = threading.Lock()
        self.writer = Serializer(sock)

    def __iter__(self) -> Self:
        return self

    def __next__(self) -> Any:
        with self.rlock:
            return next(self.reader_iter)

    def send(self, msg: Any) -> None:
        with self.wlock:
            self.writer.send(msg)



# Copyright 2015 Rackspace Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

from collections.abc import Callable
from collections.abc import Iterable
import copy
import enum
import functools
import logging
import multiprocessing
import shlex
import threading
from typing import Any

from oslo_config import cfg
from oslo_config import types
from oslo_utils import importutils

from oslo_privsep._i18n import _
from oslo_privsep import capabilities
from oslo_privsep import daemon


LOG = logging.getLogger(__name__)


def CapNameOrInt(value: str | int) -> int:
    value_str = str(value).strip()
    try:
        return capabilities.CAPS_BYNAME[value_str]
    except KeyError:
        return int(value_str)


OPTS = [
    cfg.StrOpt('user', help=_('User that the privsep daemon should run as.')),
    cfg.StrOpt(
        'group', help=_('Group that the privsep daemon should run as.')
    ),
    cfg.Opt(
        'capabilities',
        type=types.List(CapNameOrInt),
        default=[],
        help=_('List of Linux capabilities retained by the privsep daemon.'),
    ),
    cfg.IntOpt(
        'thread_pool_size',
        min=1,
        help=_(
            "The number of threads available for privsep to "
            "concurrently run processes. Defaults to the number of "
            "CPU cores in the system."
        ),
        default=multiprocessing.cpu_count(),
        sample_default='multiprocessing.cpu_count()',
    ),
    cfg.StrOpt(
        'helper_command',
        help=_(
            'Command to invoke to start the privsep daemon if '
            'not using the "fork" method. '
            'If not specified, a default is generated using '
            '"sudo privsep-helper" and arguments designed to '
            'recreate the current configuration. '
            'This command must accept suitable --privsep_context '
            'and --privsep_sock_path arguments.'
        ),
    ),
    cfg.StrOpt(
        'logger_name',
        help=_(
            'Logger name to use for this privsep context.  By '
            'default all contexts log with oslo_privsep.daemon.'
        ),
        default='oslo_privsep.daemon',
    ),
    cfg.BoolOpt(
        'log_daemon_traceback',
        help=_(
            'Print the exception traceback happened in the daemon '
            'in the client logger'
        ),
        default=False,
    ),
]

_ENTRYPOINT_ATTR = 'privsep_entrypoint'
_HELPER_COMMAND_PREFIX = ['sudo']


def _list_opts() -> list[tuple[cfg.OptGroup, list[cfg.Opt]]]:
    """Returns a list of oslo.config options available in the library.

    The returned list includes all oslo.config options which may be registered
    at runtime by the library.

    Each element of the list is a tuple. The first element is the name of the
    group under which the list of elements in the second element will be
    registered. A group name of None corresponds to the [DEFAULT] group in
    config files.

    The purpose of this is to allow tools like the Oslo sample config file
    generator to discover the options exposed to users by this library.

    :returns: a list of (group_name, opts) tuples
    """
    # This is the default group name, but that can be overridden by the caller
    group = cfg.OptGroup(
        'privsep',
        title='oslo.privsep options',
        help='Configuration options for the oslo.privsep '
        'daemon. Note that this group name can be '
        'changed by the consuming service. Check the '
        'service\'s docs to see if this is the case.',
    )
    return [(group, copy.deepcopy(OPTS))]


@enum.unique
class Method(enum.Enum):
    FORK = 1
    ROOTWRAP = 2


def init(root_helper: list[str] | None = None) -> None:
    """Initialise oslo.privsep library.

    This function should be called at the top of main(), after the
    command line is parsed, oslo.config is initialised and logging is
    set up, but before calling any privileged entrypoint, changing
    user id, forking, or anything else "odd".

    :param root_helper: List of command and arguments to prefix
        privsep-helper with, in order to run helper as root.  Note,
        ignored if context's helper_command config option is set.
    """

    if root_helper:
        global _HELPER_COMMAND_PREFIX
        _HELPER_COMMAND_PREFIX = root_helper


class PrivContext:
    def __init__(
        self,
        prefix: str,
        cfg_section: str = 'privsep',
        pypath: str | None = None,
        capabilities: Iterable[int] | None = None,
        logger_name: str = 'oslo_privsep.daemon',
        timeout: float | None = None,
    ) -> None:
        # Note that capabilities=[] means retaining no capabilities
        # and leaves even uid=0 with no powers except being able to
        # read/write to the filesystem as uid=0.  This might be what
        # you want, but probably isn't.
        #
        # There is intentionally no way to say "I want all the
        # capabilities."
        if capabilities is None:
            raise ValueError('capabilities is a required parameter')

        self.pypath = pypath
        self.prefix = prefix
        self.cfg_section = cfg_section

        self.client_mode = True
        self.channel: daemon._ClientChannel | None = None
        self.start_lock = threading.Lock()

        cfg.CONF.register_opts(OPTS, group=cfg_section)
        cfg.CONF.set_default(
            'capabilities', group=cfg_section, default=list(capabilities)
        )
        cfg.CONF.set_default(
            'logger_name', group=cfg_section, default=logger_name
        )
        self.timeout = timeout

    @property
    def conf(self) -> Any:
        """Return the oslo.config section object as lazily as possible."""
        # Need to avoid looking this up before oslo_config has been
        # properly initialized.
        return cfg.CONF[self.cfg_section]

    def __repr__(self) -> str:
        return f'PrivContext(cfg_section={self.cfg_section})'

    def helper_command(self, sockpath: str) -> list[str]:
        # We need to be able to reconstruct the context object in the new
        # python process we'll get after rootwrap/sudo.  This means we
        # need to construct the context object and store it somewhere
        # globally accessible, and then use that python name to find it
        # again in the new python interpreter.  Yes, it's all a bit
        # clumsy, and none of it is required when using the fork-based
        # alternative above.
        # These asserts here are just attempts to catch errors earlier.
        # TODO(gus): Consider replacing with setuptools entry_points.
        if self.pypath is None:
            raise AssertionError(
                'helper_command requires priv_context pypath to be specified'
            )
        if importutils.import_class(self.pypath) is not self:
            raise AssertionError(
                'helper_command requires priv_context '
                'pypath for context object'
            )

        # Note order is important here.  Deployments will (hopefully)
        # have the exact arguments in sudoers/rootwrap configs and
        # reordering args will break configs!

        if self.conf.helper_command:
            cmd = shlex.split(self.conf.helper_command)
        else:
            cmd = _HELPER_COMMAND_PREFIX + ['privsep-helper']

            try:
                for cfg_file in cfg.CONF.config_file:
                    cmd.extend(['--config-file', cfg_file])
            except cfg.NoSuchOptError:
                pass

            try:
                if cfg.CONF.config_dir is not None:
                    for cfg_dir in cfg.CONF.config_dir:
                        cmd.extend(['--config-dir', cfg_dir])
            except cfg.NoSuchOptError:
                pass

        cmd.extend(
            ['--privsep_context', self.pypath, '--privsep_sock_path', sockpath]
        )

        return cmd

    def set_client_mode(self, enabled: bool) -> None:
        self.client_mode = enabled

    def entrypoint(self, func: Callable[..., Any]) -> functools.partial[Any]:
        """This is intended to be used as a decorator."""
        return self._entrypoint(func)

    def entrypoint_with_timeout(
        self, timeout: float
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        """This is intended to be used as a decorator with timeout."""

        def wrap(func: Callable[..., Any]) -> Callable[..., Any]:
            @functools.wraps(func)
            def inner(*args: Any, **kwargs: Any) -> Any:
                f = self._entrypoint(func)
                return f(*args, _wrap_timeout=timeout, **kwargs)

            setattr(inner, _ENTRYPOINT_ATTR, self)
            return inner

        return wrap

    def _entrypoint(self, func: Callable[..., Any]) -> functools.partial[Any]:
        if not func.__module__.startswith(self.prefix):
            raise AssertionError(
                f'{self!r} entrypoints must be below "{self.prefix}"'
            )

        # Right now, we only track a single context in
        # _ENTRYPOINT_ATTR.  This could easily be expanded into a set,
        # but that will increase the memory overhead.  Revisit if/when
        # someone has a need to associate the same entrypoint with
        # multiple contexts.
        if getattr(func, _ENTRYPOINT_ATTR, None) is not None:
            raise AssertionError(
                f'{func!r} is already associated with another PrivContext'
            )

        f = functools.partial(self._wrap, func)
        setattr(f, _ENTRYPOINT_ATTR, self)
        return f

    def is_entrypoint(self, func: Callable[..., Any]) -> bool:
        return getattr(func, _ENTRYPOINT_ATTR, None) is self

    def _wrap(
        self,
        func: Callable[..., Any],
        *args: Any,
        _wrap_timeout: float | None = None,
        **kwargs: Any,
    ) -> Any:
        if self.client_mode:
            name = f'{func.__module__}.{func.__name__}'
            if self.channel is not None and not self.channel.running:
                LOG.warning("RESTARTING PrivContext for %s", name)
                self.stop()
            if self.channel is None:
                self.start()
            if self.channel is None:
                # narrow type: this will always be non-None thank to the above
                raise RuntimeError('channel is not initialized')
            r_call_timeout = _wrap_timeout or self.timeout
            return self.channel.remote_call(name, args, kwargs, r_call_timeout)
        else:
            return func(*args, **kwargs)

    def start(self, method: Method = Method.ROOTWRAP) -> None:
        with self.start_lock:
            if self.channel is not None:
                LOG.warning('privsep daemon already running')
                return

            channel: daemon._ClientChannel
            if method is Method.ROOTWRAP:
                channel = daemon.RootwrapClientChannel(context=self)
            elif method is Method.FORK:
                channel = daemon.ForkingClientChannel(context=self)
            else:
                raise ValueError(f'Unknown method: {method}')

            self.channel = channel

    def stop(self) -> None:
        if self.channel is not None:
            self.channel.close()
            self.channel = None
